/* Copyright 1998-2011 Research Foundation State University of New York */

/* This file is part of QuB.                                            */

/* QuB is free software; you can redistribute it and/or modify          */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation, either version 3 of the License, or    */
/* (at your option) any later version.                                  */

/* QuB is distributed in the hope that it will be useful,               */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the        */
/* GNU General Public License for more details.                         */

/* You should have received a copy of the GNU General Public License,   */
/* named LICENSE.txt, in the QuB program directory.  If not, see        */
/* <http://www.gnu.org/licenses/>.                                      */

#include "qubopt.h"

#include "QUB_IdlSeg.h"
using namespace std;

using namespace std;



inline int WhichMap(int first)
{
	return first >> QUB_IDLSEG_CHUNKSHIFT;// / QUB_IDLSEG_CHUNKSIZE;
}

inline int HowManyMaps(int count)
{
	if ( count <= 0 )
		return 0;
	else
		return (count >> QUB_IDLSEG_CHUNKSHIFT) + ((count % QUB_IDLSEG_CHUNKSIZE) ? 1 : 0);
}



QUB_IdlChunk::QUB_IdlChunk(int afirst, QUB_NodeStash& astash)
	: first( afirst ), stash( astash ), storage( QUB_FS_0 ), stashKey( -1 )
{
	asMap = new QUB_IdlMap;
}

QUB_IdlChunk::~QUB_IdlChunk()
{
	if ( asMap )
		delete asMap;
	if ( stashKey >= 0 )
		stash.replace( stashKey );
}

int QUB_IdlChunk::size()
{
	if ( asMap )
		return asMap->size();
	if ( stashKey >= 0 )
		return asNode.dataRows();
	return 0;
}

void QUB_IdlChunk::setupStorage(int lvl)
{
	switch ( lvl )
	{
	case 0:
		if ( ! asMap ) {
			asMap = new QUB_IdlMap;
			asNode.loadData();
			QUB_IdlDwell *dwells = (QUB_IdlDwell *) asNode.data();
			for ( int i = asNode.dataRows() - 1; i >= 0; --i )
				(*asMap)[ dwells[i].first ] = dwells[i];
		}
		storage |= QUB_FS_0;
		break;
	case 1:
		if ( stashKey < 0 ) {
			stashKey = stash.request();
			asNode = stash.get( stashKey );
			asNode.setNumData(QTR_TYPE_INT, size(), 3, (void *) 0);
			QUB_IdlDwell *dwells = (QUB_IdlDwell *) asNode.data();
			for ( QUB_IdlMap::iterator di = asMap->begin(); di != asMap->end(); ++di )
				*(dwells++) = di->second;
		}
		asNode.loadData();
		storage |= QUB_FS_1;
		break;
	case 2:
		if ( stashKey < 0 ) {
			stashKey = stash.request();
			asNode = stash.get( stashKey );
			asNode.setNumData(QTR_TYPE_INT, size(), 3, (void *) 0);
			QUB_IdlDwell *dwells = (QUB_IdlDwell *) asNode.data();
			for ( QUB_IdlMap::iterator di = asMap->begin(); di != asMap->end(); ++di )
				*(dwells++) = di->second;
		}
		asNode.unloadData();
		storage |= QUB_FS_2;
		break;
	}
}

void QUB_IdlChunk::clearStorage(int lvl)
{
	switch ( lvl )
	{
	case 0:
		if ( asMap ) {
			delete asMap;
			asMap = NULL;
		}
		storage &= ~ QUB_FS_0;
		break;
	case 1:
		if ( stashKey >= 0 ) {
			if ( ! (storage & QUB_FS_2) ) {
				asNode.clearData();
				asNode = QUB_Tree();
				stash.replace( stashKey );
				stashKey = -1;
			}
			else
				asNode.unloadData();
		}
		storage &= ~ QUB_FS_1;
		break;
	case 2:
		storage &= ~ QUB_FS_2;
		break;
	}
}

int QUB_IdlChunk::getStorage()
{
	return storage;
}

void QUB_IdlChunk::clear(int f, int l)
{
	if ( ! asMap ) {
		if ( (f <= first) && (l >= (first + QUB_IDLSEG_CHUNKSIZE - 1)) ) {
			asNode.clearData();
			return;
		}
		else {
			setupStorage(0);
		}
	}

	if ( ! asMap->size() )
		return;

	f = max(f, first);
	l = min(l, first + QUB_IDLSEG_CHUNKSIZE - 1);

	QUB_IdlMap::iterator di = asMap->upper_bound(f);
	--di;
	if ( di->second.last < f )
		return;

	if ( di->second.first < f ) {
		if ( l < di->second.last ) {
			QUB_IdlDwell& dwell = (*asMap)[l+1];
			dwell.first = l+1;
			dwell.last = di->second.last;
			dwell.cls = di->second.cls;
		}
		di->second.last = f-1;
		++di;
	}
	while ( di != asMap->end() && di->second.last <= l ) {
		QUB_IdlMap::iterator toErase = di;
		++di;
		asMap->erase(toErase);
	}
	if ( di != asMap->end() && di->second.first <= l ) {
		QUB_IdlDwell& dwell = (*asMap)[l+1];
		dwell.first = l+1;
		dwell.last = di->second.last;
		dwell.cls = di->second.cls;
		asMap->erase(di);
	}
}

void QUB_IdlChunk::set(int f, int l, int c, bool atBegin, bool atEnd) // assuming already cleared
{
	if ( ! asMap )
		return;

	f = max(f, first);
	l = min(l, first + QUB_IDLSEG_CHUNKSIZE - 1 );

	QUB_IdlDwell *dwell = & (*asMap)[f];
	dwell->first = f;
	dwell->last = l;
	dwell->cls = c;

	if ( /*atBegin &&*/ f > first ) {
		QUB_IdlMap::iterator di = asMap->upper_bound(f-1);
		if ( di != asMap->begin() ) {
			--di;
			if ( di->second.cls == c && (di->second.last == (f-1)) ) {
				di->second.last = l;
				asMap->erase( asMap->find(f) );
				dwell = & di->second;
			}
		}
	}
	if ( /*atEnd &&*/ l < (first + QUB_IDLSEG_CHUNKSIZE - 1) ) {
		QUB_IdlMap::iterator di = asMap->upper_bound(l+1);
		--di;
		if ( di->second.cls == c && (di->first == (l+1)) ) {
			dwell->last = di->second.last;
			asMap->erase(di);
		}
	}
}

void QUB_IdlChunk::setComplete(int count, QUB_IdlDwell *buf, int offset, int pointsInSeg)
{
	if ( asMap )
		asMap->clear();
	setupStorage(1);
	clearStorage(0);

	asNode.setNumData(QTR_TYPE_INT, count, 3, (void*)0);
	QUB_IdlDwell *dwells = (QUB_IdlDwell *) asNode.data();
	for ( int i=0; i<count; ++i ) {
		dwells[i].first = buf[i].first - offset;
		dwells[i].last = buf[i].last - offset;
		dwells[i].cls = buf[i].cls;;
	}
	dwells[0].first = max(first, dwells[0].first);
	dwells[count-1].last = min(dwells[count-1].last, min(first + QUB_IDLSEG_CHUNKSIZE, pointsInSeg) - 1);
}

void QUB_IdlChunk::setComplete(int count, int *ff, int *ll, int *cc, int offset, int pointsInSeg)
{
	if ( asMap )
		asMap->clear();
	setupStorage(1);
	clearStorage(0);

	asNode.setNumData(QTR_TYPE_INT, count, 3, (void*)0);
	QUB_IdlDwell *dwells = (QUB_IdlDwell *) asNode.data();
	for ( int i=0; i<count; ++i ) {
		dwells[i].first = ff[i] - offset;
		dwells[i].last = ll[i] - offset;
		dwells[i].cls = cc[i];;
	}
	dwells[0].first = max(first, dwells[0].first);
	dwells[count-1].last = min(dwells[count-1].last, min(first + QUB_IDLSEG_CHUNKSIZE, pointsInSeg) - 1);
}

int QUB_IdlChunk::set(int count, QUB_IdlDwell *buf, int offset, int pointsInSeg)
{
	int i0 = 0;
	while ( i0 < count && (buf[i0].last - offset) < first )
		++i0;
	if ( i0 == count )
		return count;

	int n = i0;
	while ( (n+1) < count && (buf[n].last - offset - first) < QUB_IDLSEG_CHUNKSIZE )
		++n;
	if ( (n+1) < count && (buf[n].first - offset - first) >= QUB_IDLSEG_CHUNKSIZE )
		--n;

	if ( (buf[i0].first - offset <= first) && (buf[n].last - offset >= (first + QUB_IDLSEG_CHUNKSIZE - 1)) ) {
		setComplete(n - i0 + 1, buf+i0, offset, pointsInSeg);
	}
	else {
		setupStorage(0);
		clear( buf[i0].first - offset, buf[n].last - offset );
		set( buf[i0].first - offset, min(buf[i0].last - offset, pointsInSeg-1), buf[i0].cls, true, i0==n );
		int i = i0 + 1;
		while ( i < n ) {
			set( buf[i].first - offset, buf[i].last - offset, buf[i].cls, false, false );
			++i;
		}
		if ( i0 != n )
			set( buf[n].first - offset,
			min(buf[n].last - offset, pointsInSeg-1), buf[n].cls, false, true );
	}
	return n + ((buf[n].last - offset - first) < QUB_IDLSEG_CHUNKSIZE);
}

int QUB_IdlChunk::set(int count, int *ff, int *ll, int *cc, int offset, int pointsInSeg)
{
	int i0 = 0;
	while ( i0 < count && (ll[i0] - offset) < first )
		++i0;
	if ( i0 == count )
		return count;

	int n = i0;
	while ( (n+1) < count && (ll[n] - offset - first) < QUB_IDLSEG_CHUNKSIZE )
		++n; // n: first event ending past this chunk
	if ( (n+1) < count && (ff[n] - offset - first) >= QUB_IDLSEG_CHUNKSIZE )
		--n; // n: last event fully/partially in chunk

	if ( (ff[i0] - offset - first <= 0) && (ll[n] - offset - first >= (QUB_IDLSEG_CHUNKSIZE - 1)) ) {
		setComplete(n - i0 + 1, ff+i0, ll+i0, cc+i0, offset, pointsInSeg);
	}
	else {
		setupStorage(0);
		clear( ff[i0] - offset, ll[n] - offset );
		set( ff[i0] - offset, min(ll[i0] - offset, pointsInSeg-1), cc[i0], true, i0==n );
		int i = i0 + 1;
		while ( i < n ) {
			set( ff[i] - offset, ll[i] - offset, cc[i], false, false );
			++i;
		}
		if ( i0 != n )
			set( ff[n] - offset, 
			min(ll[n] - offset, pointsInSeg-1), cc[n], false, true );
	}
	return n + ((ll[n] - offset - first) < QUB_IDLSEG_CHUNKSIZE);
}


QUB_IdlMaps::QUB_IdlMaps(QUB_IdlFreezer& afreezer, QUB_NodeStash& stash, int apoints)
: freezer( afreezer ), chunks( &afreezer ), chunkStorage( stash ), points( 0 )
{
	if ( apoints ) {
		setSize( apoints );
		for ( int i=0; i<=maxmap; ++i ) {
			QUB_IdlChunk *chunk = chunks.checkOut( i, 0x1 | 0x2 );
			chunk->set( i*QUB_IDLSEG_CHUNKSIZE, min(points, (i+1) * QUB_IDLSEG_CHUNKSIZE) - 1, -1, false, false );
			chunks.checkIn( i, true );
		}
	}
	else
		maxmap = -1;
}

QUB_IdlMaps::~QUB_IdlMaps()
{
}

void QUB_IdlMaps::setSize( int n )
{
	if ( n > points ) {
		maxmap = HowManyMaps(n) - 1;
		for ( int i=chunks.size(); i<=maxmap; ++i )
			chunks.insert( chunks.size(), new QUB_IdlChunk( i * QUB_IDLSEG_CHUNKSIZE, chunkStorage ) );
		points = n;
	}
	else if ( n < points ) {
		chunks.truncate( HowManyMaps(n) );
		maxmap = chunks.size() - 1;
		points = n;

		if ( points % QUB_IDLSEG_CHUNKSIZE ) {
			QUB_IdlChunk *chunk = chunks.checkOut(maxmap, 0x1 | 0x2);
			chunk->clear(points, points + QUB_IDLSEG_CHUNKSIZE);
			chunks.checkIn(maxmap, true);
		}
	}
}

void QUB_IdlMaps::insert(int f, int l)
{
	vector<QUB_IdlDwell> dwells;
	if ( true ) {
		iterator di = find(f);
		if ( di && di->first < f ) {
			dwells.resize(1);
			dwells[0].first = f;
			dwells[0].last = di->last;
			dwells[0].cls = di->cls;
			++di;
		}
		while ( di ) {
			dwells.push_back( *di );
			++di;
		}
	}

	int oldLen = points;
	int offset = l - f + 1;
	if ( f < points )
		remove(f, points-1);

	setSize( oldLen + offset );
	set(f, l, -1);
	if ( dwells.size() )
		set(dwells.size(), &(dwells[0]), - offset);
}

void QUB_IdlMaps::remove(int f, int l)
{
	vector<QUB_IdlDwell> dwells;
	if ( true ) { // make di deleted before making changes
		iterator di = find(l+1);
		if ( di && di->first <= l ) {
			dwells.resize(1);
			dwells[0].first = l+1;
			dwells[0].last = di->last;
			dwells[0].cls = di->cls;
			++di;
		}
		while ( di ) {
			dwells.push_back( *di );
			++di;
		}
	}
	
	int origPoints = points;
	int offset = (l - f + 1);
	setSize( f );
	setSize( origPoints - offset );
	if ( dwells.size() )
		set(dwells.size(), &(dwells[0]), offset);
}

void QUB_IdlMaps::set(int f, int l, int c)
{
	set(1, &f, &l, &c, 0);
}

void QUB_IdlMaps::join(int f, int l)
{
	iterator di = find(f);
	if ( di ) {
		if ( f && (di->first == f) )
			--di;
		set(f, l, di->cls);
	}
}

int QUB_IdlMaps::set(int count, QUB_IdlDwell *buf, int offset)
{
	if ( ! count )
		return 0;

	int kf = max( WhichMap(buf[0].first - offset), 0 );
	int kl = min( WhichMap(buf[count-1].last - offset), maxmap );
	int totalConsumed = 0;

	for ( int ki=kf; totalConsumed<count && ki<=kl; ++ki ) {
		QUB_IdlChunk *chunk = chunks.checkOut(ki, 0x1 | 0x2);
		int consumed = chunk->set(count - totalConsumed, buf + totalConsumed, offset, points);
		totalConsumed += consumed;
		chunks.checkIn( ki, true );
	}
	return totalConsumed;
}

int QUB_IdlMaps::set(int count, int *ff, int *ll, int *cc, int offset)
{
	if ( ! count )
		return 0;

	int kf = max( WhichMap(ff[0] - offset), 0 );
	int kl = min( WhichMap(ll[count-1] - offset), maxmap );
	int totalConsumed = 0;

	for ( int ki=kf; totalConsumed<count && ki<=kl; ++ki ) {
		QUB_IdlChunk *chunk = chunks.checkOut(ki, 0x1 | 0x2);
		int consumed = chunk->set(count - totalConsumed, ff + totalConsumed, ll + totalConsumed, cc + totalConsumed, offset, points);
		totalConsumed += consumed;
		chunks.checkIn( ki, true );
	}
	return totalConsumed;
}

QUB_IdlMapsIter::QUB_IdlMapsIter(QUB_IdlMaps& amaps, int point)
: maps( &amaps )
{
	if ( point < 0 || point >= maps->points ) {
		ix = maps->maxmap;
		if ( ix >= 0 ) {
			chunk = maps->chunks.checkOut( ix, 0x1 | 0x2 );
			if ( chunk->getStorage() & 0x1 ) {
				iter = chunk->asMap->end();
				arr = NULL;
			}
			else {
				arr = (QUB_IdlDwell *) chunk->asNode.data();
				count = chunk->size();
				i = count;
			}
		}
		else {
			chunk = NULL;
		}
		return;
	}

	ix = WhichMap(point);
	chunk = maps->chunks.checkOut( ix, 0x1 | 0x2 );
	if ( chunk->getStorage() & QUB_FS_0 ) {
		iter = chunk->asMap->upper_bound(point);
		if ( chunk->asMap->size() )
			--iter;
		arr = NULL;
	}
	else {
		arr = (QUB_IdlDwell *) chunk->asNode.data();
		count = chunk->size();
		for ( i=0; i < count && arr[i].last < point; ++i )
			;
	}
}

QUB_IdlMapsIter::QUB_IdlMapsIter(QUB_IdlMaps& amaps, bool atBegin)
: maps( &amaps )
{
	if ( maps->maxmap < 0 ) {
		ix = maps->maxmap;
		chunk = NULL;
	}
	else {
		if ( atBegin ) {
			ix = 0;
			chunk = maps->chunks.checkOut( ix, 0x1 | 0x2 );
			if ( chunk->getStorage() & QUB_FS_0 ) {
				iter = chunk->asMap->begin();
				arr = NULL;
			}
			else {
				arr = (QUB_IdlDwell *) chunk->asNode.data();
				count = chunk->size();
				i = 0;
			}
		}
		else {
			ix = maps->maxmap;
			chunk = maps->chunks.checkOut( ix, 0x1 | 0x2 );
			if ( chunk->getStorage() & QUB_FS_0 ) {
				iter = chunk->asMap->end();
				arr = NULL;
			}
			else {
				arr = (QUB_IdlDwell *) chunk->asNode.data();
				count = chunk->size();
				i = count;
			}
		}
	}
}

QUB_IdlMapsIter::~QUB_IdlMapsIter()
{
	if ( ix >= 0 )
		maps->chunks.checkIn( ix, false );
}

QUB_IdlMapsIter& QUB_IdlMapsIter::operator=(const QUB_IdlMapsIter& other)
{
	if ( maps != other.maps || ix != other.ix ) {
		if ( ix >= 0 )
			maps->chunks.checkIn( ix, false );
		maps = other.maps;
		ix = other.ix;
		if ( ix >= 0 )
			chunk = maps->chunks.checkOut( ix, 0x1 | 0x2 );
		else
			chunk = NULL;
	}
	iter = other.iter;
	arr = other.arr;
	count = other.count;
	i = other.i;

	return *this;
}

QUB_IdlMapsIter& QUB_IdlMapsIter::operator++ ()
{
	if ( ix < 0 || (arr ? (i == count) : (iter == chunk->asMap->end())) )
		return *this;

	bool atEnd;
	if ( arr ) {
		++i;
		atEnd = i == count;
	}
	else {
		++iter;
		atEnd = iter == chunk->asMap->end();
	}
	if ( atEnd && ix < maps->maxmap ) {
		maps->chunks.checkIn( ix, false );

		++ix;
		chunk = maps->chunks.checkOut( ix, 0x1 | 0x2 );
		if ( chunk->getStorage() & QUB_FS_0 ) {
			iter = chunk->asMap->begin();
			arr = NULL;
		}
		else {
			arr = (QUB_IdlDwell *) chunk->asNode.data();
			count = chunk->size();
			i = 0;
		}
	}
	return *this;
}

QUB_IdlMapsIter& QUB_IdlMapsIter::operator-- ()
{
	if ( arr ? i : iter != chunk->asMap->begin() ) {
		if ( arr )
			--i;
		else
			--iter;
	}
	else if ( ix ) {
		maps->chunks.checkIn( ix, false );

		--ix;
		chunk = maps->chunks.checkOut( ix, 0x1 | 0x2 );
		if ( chunk->getStorage() & QUB_FS_0 ) {
			iter = chunk->asMap->end();
			--iter;
			arr = NULL;
		}
		else {
			arr = (QUB_IdlDwell *) chunk->asNode.data();
			count = chunk->size();
			i = count - 1;
		}
	}
	return *this;
}


QUB_IdlSeg::QUB_IdlSeg(QUB_IdlFreezer& afreezer, QUB_NodeStash& stash, int afirst, int alast, double samp)
: freezer( afreezer ), chunkStorage( stash ), sampling( samp ), first( afirst ), last( alast ), maps( afreezer, stash, alast - afirst + 1 )
{
}

void      QUB_IdlSeg::setFirst(int f)
{
	if ( f > first ) {
		if ( f > last )
			f = last + 1;
		maps.remove(0, (f-first)-1);
	}
	else if ( f < first ) {
		maps.insert(0, (first - f)-1);
	}
	first = f;
}

void      QUB_IdlSeg::setLast(int l)
{
	if ( l < last ) {
		if ( l < first )
			l = first - 1;
		maps.remove(l - first + 1, last - first);
	}
	else if ( l > last ) {
		maps.insert(last - first + 1, l - first);
	}
	last = l;
}

void      QUB_IdlSeg::offset(int points)
{
	first += points;
	last += points;
}

void      QUB_IdlSeg::insert(int f, int l)
{
	maps.insert(f-first, l-first);
	last += (l - f + 1);
}

void      QUB_IdlSeg::remove(int f, int l)
{
	maps.remove(f-first, l-first);
	last -= (l - f + 1);
}

void      QUB_IdlSeg::set(int f, int l, int c)
{
	maps.set(f-first, l-first, c);
}

void      QUB_IdlSeg::join(int f, int l)
{
	maps.join(f-first, l-first);
}

// this will move begin up to end
QUB_Tree BuildSegNode( int count, QUB_IdlMaps::iterator& begin, QUB_IdlMaps::iterator& end, int offset, int f, int l,
				   QUB_Tree amp, QUB_Tree sd, double sampling, int withDurs, int dropF, int dropL, int dropIfCls, int dropCls )
{
	QUB_Tree seg = QUB_Tree::Create("Segment");
	seg.setNumData(QTR_TYPE_INT, 2, 1, (void*)0);

	QUB_IdlMaps::iterator actualEnd = end;

	int lastCls;
	if ( dropIfCls ) {
		dropF = begin->cls == dropCls;
		--actualEnd;
		dropL = actualEnd->cls == dropCls;
		++actualEnd;
	}

	count -= dropF + dropL;
	if ( count <= 0 ) {
		while ( begin != end )
			++begin;
		return QUB_Tree();
	}

	if ( dropF ) {
		lastCls = begin->cls;
		while ( begin->cls == lastCls )
			++begin;
	}
	if ( dropL ) {
		--actualEnd;
		lastCls = actualEnd->cls;
		while ( actualEnd->cls == lastCls )
			--actualEnd;
		++actualEnd; // one-past, stl-style
	}

	seg["DwellCount"].setData(QTR_TYPE_INT, count);

	QUB_Tree firsts  = seg["Firsts"];
	QUB_Tree lasts   = seg["Lasts"];
	QUB_Tree classes = seg["Classes"];
	firsts.setNumData(QTR_TYPE_INT, count, 1, (void *)0);
	lasts.setNumData(QTR_TYPE_INT, count, 1, (void *)0);
	classes.setNumData(QTR_TYPE_INT, count, 1, (void *)0);

	int *ff = (int *) firsts.data();
	int *ll = (int *) lasts.data();
	int *cc = (int *) classes.data();
	int i = -1;
	lastCls = -1;

	while ( begin != actualEnd ) {
		if ( begin->cls == lastCls ) {
			ll[i] = begin->last + offset;
		}
		else {
			lastCls = begin->cls;
			++i;
			ff[i] = begin->first + offset;
			ll[i] = begin->last + offset;
			cc[i] = lastCls;
		}
		++begin;
	}
	seg.dataAs(0, 0, 0) = (ff[0] = max(ff[0], f+offset));
	seg.dataAs(1, 0, 0) = (ll[count-1] = min(ll[count-1], l+offset));

	while ( begin != end )
		++begin;

	if ( withDurs ) {
		sampling *= 1.0e3;
		QUB_Tree durs    = seg["Durations"];
		durs.setNumData(QTR_TYPE_FLOAT, count, 1, (void *)0);
		float *tdwt = (float *) durs.data();
		for ( int d=0; d<count; ++d )
			tdwt[d] = float(sampling * (ll[d] - ff[d] + 1));
	}

	seg.insertClone(sd);
	seg.insertClone(amp);

	return seg;
}

void BuildUpSegNode(QUB_Tree seg, int count, QUB_IdlMaps::iterator& begin, QUB_IdlMaps::iterator& end,
					int offset, int f, int l, double sampling, int withDurs)
{
	seg.setNumData(QTR_TYPE_INT, 2, 1, (void*)0);

	seg["DwellCount"].setData(QTR_TYPE_INT, count);

	QUB_Tree firsts  = seg["Firsts"];
	QUB_Tree lasts   = seg["Lasts"];
	QUB_Tree classes = seg["Classes"];
	firsts.setNumData(QTR_TYPE_INT, count, 1, (void *)0);
	lasts.setNumData(QTR_TYPE_INT, count, 1, (void *)0);
	classes.setNumData(QTR_TYPE_INT, count, 1, (void *)0);

	int *ff = (int *) firsts.data();
	int *ll = (int *) lasts.data();
	int *cc = (int *) classes.data();
	int i = -1;
	int lastCls = -1;

	while ( begin != end ) {
		if ( begin->cls == lastCls ) {
			ll[i] = begin->last + offset;
		}
		else {
			lastCls = begin->cls;
			++i;
			ff[i] = begin->first + offset;
			ll[i] = begin->last + offset;
			cc[i] = lastCls;
		}
		++begin;
	}
	seg.dataAs(0, 0, 0) = (ff[0] = max(ff[0], f+offset));
	seg.dataAs(1, 0, 0) = (ll[count-1] = min(ll[count-1], l+offset));

	if ( withDurs ) {
		sampling *= 1.0e3;
		QUB_Tree durs    = seg["Durations"];
		durs.setNumData(QTR_TYPE_FLOAT, count, 1, (void *)0);
		float *tdwt = (float *) durs.data();
		for ( int d=0; d<count; ++d )
			tdwt[d] = float(sampling * (ll[d] - ff[d] + 1));
	}
}

void QUB_IdlSeg::getDwells(int f, int l, int fragments, int withDurs, QUB_Tree ds, QUB_Tree& afterSeg)
{
	f = max(f, first) - first;
	l = min(l, last)  - first;

	double samp = sampling * 1.0e3;
	if ( ds.find("sampling")->isNull() ) {
		ds["sampling"].setData(QTR_TYPE_DOUBLE, samp);
	}

	QUB_Tree amp = QUB_Tree::Create("amp");
	QUB_Tree sd  = QUB_Tree::Create("sd");
	int nclass = getClassCount();
	if ( nclass ) {
		amp.setNumData(QTR_TYPE_DOUBLE, nclass, 1, (void *)0);
		sd.setNumData(QTR_TYPE_DOUBLE, nclass, 1, (void *)0);
		for ( int c=0; c<nclass; ++c ) {
			amp.dataAs( c, 0, (double) 0.0 ) = getAmp(c);
			sd.dataAs( c, 0, (double) 0.0 ) =  getStd(c);
		}
	}

	QUB_IdlMaps::iterator counter, reader;
	int count = 0;
	int lastCls = -1;
	int lastPt;

	counter = find(f+first);
	if ( (! fragments) && (counter->first < f) ) {
		lastCls = counter->cls;
		do {
			++counter;
		} while ( counter && (counter->cls == lastCls) );
		lastCls = -1;
	}

	while ( counter && (counter->first <= l) ) {
		if ( counter->cls < 0 ) {
			lastCls = -1;
			if ( count ) {
				QUB_Tree thisseg = BuildSegNode( count, reader, counter, first, f, l, amp, sd, sampling, withDurs, 0, fragments ? 0 : (lastPt != l), 0, 0 );
				if ( ! thisseg.isNull() )
					afterSeg = ds.insertChild( afterSeg, thisseg );
				count = 0;
			}
		}
		else {
			if ( count == 0 )
				reader = counter;
			if ( counter->cls != lastCls )
				++count;
			lastCls = counter->cls;
			lastPt  = counter->last;
		}
		++counter;
	}

	if ( count ) {
		QUB_Tree thisseg = BuildSegNode( count, reader, counter, first, f, l, amp, sd, sampling, withDurs, 0, fragments ? 0 : (lastPt != l), 0, 0 );
		if ( ! thisseg.isNull() )
			afterSeg = ds.insertChild( afterSeg, thisseg );
	}
}

// the assumption is that you drop first/last whether or not they're fragmental events

void      QUB_IdlSeg::getDwellsWDrop(int f, int l, int withDurs, int dropFL, int dropIfCls, int dropCls, QUB_Tree dataSet, QUB_Tree& afterSeg )
{
	f = max(f, first) - first;
	l = min(l, last)  - first;

	double samp = sampling * 1.0e3;
	if ( dataSet.find("sampling")->isNull() ) {
		dataSet["sampling"].setData(QTR_TYPE_DOUBLE, samp);
	}

	QUB_Tree amp = QUB_Tree::Create("amp");
	QUB_Tree sd  = QUB_Tree::Create("sd");
	int nclass = getClassCount();
	if ( nclass ) {
		amp.setNumData(QTR_TYPE_DOUBLE, nclass, 1, (void *)0);
		sd.setNumData(QTR_TYPE_DOUBLE, nclass, 1, (void *)0);
		for ( int c=0; c<nclass; ++c ) {
			amp.dataAs( c, 0, (double) 0.0 ) = getAmp(c);
			sd.dataAs( c, 0, (double) 0.0 ) =  getStd(c);
		}
	}

	QUB_IdlMaps::iterator counter, reader;
	int count = 0;
	int lastCls = -1;

	counter = find(f+first);

	while ( counter && (counter->first <= l) ) {
		if ( counter->cls < 0 ) {
			lastCls = -1;
			if ( count ) {
				QUB_Tree thisseg = BuildSegNode( count, reader, counter, first, f, l, amp, sd, sampling, withDurs, dropFL, dropFL, dropIfCls, dropCls );
				if ( ! thisseg.isNull() )
					afterSeg = dataSet.insertChild( afterSeg, thisseg );
				count = 0;
			}
		}
		else {
			if ( count == 0 )
				reader = counter;
			if ( counter->cls != lastCls )
				++count;
			lastCls = counter->cls;
		}
		++counter;
	}

	if ( count ) {
		QUB_Tree thisseg = BuildSegNode( count, reader, counter, first, f, l, amp, sd, sampling, withDurs, dropFL, dropFL, dropIfCls, dropCls );
		if ( ! thisseg.isNull() )
			afterSeg = dataSet.insertChild( afterSeg, thisseg );
	}
}

void QUB_IdlSeg::getDwellsInto(QUB_Tree seg, int withDurs)
{
	int f = max(seg.dataAs(0, 0, 0), first) - first;
	int l = min(seg.dataAs(1, 0, 0), last)  - first;

	double samp = sampling * 1.0e3;

	QUB_Tree amp = seg["amp"];
	QUB_Tree sd  = seg["sd"];
	int nclass = getClassCount();
	if ( nclass ) {
		amp.setNumData(QTR_TYPE_DOUBLE, nclass, 1, (void *)0);
		sd.setNumData(QTR_TYPE_DOUBLE, nclass, 1, (void *)0);
		for ( int c=0; c<nclass; ++c ) {
			amp.dataAs( c, 0, (double) 0.0 ) = getAmp(c);
			sd.dataAs( c, 0, (double) 0.0 ) =  getStd(c);
		}
	}

	QUB_IdlMaps::iterator counter, reader;
	int count = 0;
	int lastCls = -1;
	int lastPt;

	counter = find(f+first);
	while ( counter && (counter->first <= l) ) {
		if ( counter->cls < 0 ) {
			lastCls = -1;
			if ( count ) {
				BuildUpSegNode(seg, count, reader, counter, first, f, l, sampling, withDurs);
				count = 0;
				return; // skip the rest if not contiguous
			}
		}
		else {
			if ( count == 0 )
				reader = counter;
			if ( counter->cls != lastCls )
				++count;
			lastCls = counter->cls;
			lastPt  = counter->last;
		}
		++counter;
	}

	if ( count )
		BuildUpSegNode(seg, count, reader, counter, first, f, l, sampling, withDurs);
}

void      QUB_IdlSeg::setDwells(QUB_Tree segment)
{
	setDwells( segment["DwellCount"].dataAsInt(),
			(int *) segment["Firsts"].data(),
			(int *) segment["Lasts"].data(),
			(int *) segment["Classes"].data() );

	std::set<int> classesPresent;
	int *classes = (int *) segment["Classes"].data();
	for ( int d=segment["DwellCount"].dataAsInt()-1; d>=0; --d )
		classesPresent.insert( classes[d] );

	double *amp = (double *) segment["amp"].data();
	double *std = (double *) segment["sd"].data();
	if ( amp && std ) {
		for ( std::set<int>::iterator ci=classesPresent.begin(); ci != classesPresent.end(); ++ci ) {
			int i = *ci;
			setAmp(i, amp[i]);
			setStd(i, std[i]);
		}
	}
}

void      QUB_IdlSeg::setDwells(int count, QUB_IdlDwell *buf)
{
	int a = 0;
	while ( a<count && buf[a].first < first )
		++a;
	int b = count-1;
	while ( b>=a && buf[b].last > last )
		--b;

	maps.set(b - a + 1, buf+a, first);
}

void      QUB_IdlSeg::setDwells(int count, int *ff, int *ll, int *cc)
{
	if ( ! (ff && ll && cc) )
		return;

	int a = 0;
	while ( a<count && ff[a] < first )
		++a;
	int b = count-1;
	while ( b>=a && ll[b] > last )
		--b;

	maps.set(b - a + 1, ff+a, ll+a, cc+a, first);
}

int       QUB_IdlSeg::getClassCount()
{
	return ampStd.size();
}

void      QUB_IdlSeg::setClassCount(int cc)
{
	if ( int(ampStd.size()) > cc )
		ampStd.resize(cc);
	while ( int(ampStd.size()) < cc )
		ampStd.push_back( pair<float, float>( float(ampStd.size()), 0.1f ) );
}

double    QUB_IdlSeg::getAmp(int cls)
{
	if ( cls >= int(ampStd.size()) )
		setClassCount(cls+1);
	return ampStd[cls].first;
}

void      QUB_IdlSeg::setAmp(int cls, double amp)
{
	if ( cls >= int(ampStd.size()) )
		setClassCount(cls+1);
	ampStd[cls].first = float(amp);
}

double    QUB_IdlSeg::getStd(int cls)
{
	if ( cls >= int(ampStd.size()) )
		setClassCount(cls+1);
	return ampStd[cls].second;
}

void      QUB_IdlSeg::setStd(int cls, double std)
{
	if ( cls >= int(ampStd.size()) )
		setClassCount(cls+1);
	ampStd[cls].second = float(std);
}


QUB_IdlSegmentsIter& QUB_IdlSegmentsIter::operator++ ()
{
	if ( *this ) {
		++iter;
		if ( (seg_ix+1) < int(segs.segs.size()) && ! iter ) {
			++seg_ix;
			seg = segs.segs[seg_ix];
			iter = seg->begin();
		}
	}
	return *this;
}

QUB_IdlSegmentsIter& QUB_IdlSegmentsIter::operator-- ()
{
	if ( seg ) {
		if ( seg_ix > 0 && iter == seg->begin() ) {
			--seg_ix;
			seg = segs.segs[ seg_ix ];
			iter = seg->end();
		}
		--iter;
	}
	return *this;
}


void __cdecl QUB_IdlSegmentsIter::Free()
{
	delete this;
}

int __cdecl QUB_IdlSegmentsIter::valid()
{
	return *this ? 1 : 0;
}

void __cdecl QUB_IdlSegmentsIter::read( int *f, int *l, int *c, double *a )
{
	QUB_IdlDwell& dwell = **this;
	*f = dwell.first + seg->first;
	*l = dwell.last + seg->first;
	*c = dwell.cls;
	if ( *c >= 0 )
		*a = seg->getAmp(*c);
}

void __cdecl QUB_IdlSegmentsIter::next()
{
	++(*this);
}

void __cdecl QUB_IdlSegmentsIter::prev()
{
	--(*this);
}

QUB_IdlSegmentsIter_iface* __cdecl QUB_IdlSegmentsIter::copy()
{
	return new QUB_IdlSegmentsIter(segs, seg_ix, iter);
}


QUB_IdlSegments::QUB_IdlSegments(double samp)
: sampling(samp)
{
	InitializeCriticalSection(&crit);
	freezer.setThreshold(0, 50000);
	freezer.setThreshold(1, 500000);
}

QUB_IdlSegments::~QUB_IdlSegments()
{
	for ( int i=0; i<int(segs.size()); ++i )
		delete segs[i];
	DeleteCriticalSection(&crit);
}

void __cdecl QUB_IdlSegments::Free()
{
	delete this;
}

void __cdecl QUB_IdlSegments::lock()
{
	EnterCriticalSection(&crit);
}

void __cdecl QUB_IdlSegments::unlock()
{
	LeaveCriticalSection(&crit);
}

void       QUB_IdlSegments::getSegRange(int f, int l, int& sf, int& sl)
{
	int nseg = segs.size();
	if ( nseg == 0 ) {
		sf = 0; sl = -1; return;
	}
	if ( segs[0]->last < segs[0]->first ) {
		sf = 0; sl = 0; return;
	}

	sf = 0;
	while ( (sf < nseg) && (segs[sf]->last < f) )
		++sf;
	sl = sf;
	while ( (sl < nseg) && (segs[sl]->last < l) )
		++sl;
	if ( (sf < nseg) && (segs[sf]->last < f) )
		++sf;

	// Chris 12-2005 this is making weird stuff happen
	//if ( (sf == nseg) && (segs[nseg-1]->last == (f-1)) )
	//	sf = sl = nseg-1;
}

void       QUB_IdlSegments::resetSeg(int s)
{
	int f = segs[s]->first;
	int l = segs[s]->last;
	delete segs[s];
	segs[s] = new QUB_IdlSeg(freezer, chunkStorage, f, l, sampling);
}

double     __cdecl QUB_IdlSegments::getSampling()
{
	return sampling;
}

void       __cdecl QUB_IdlSegments::setSampling(double samp)
{
	sampling = samp;
	for ( int i=0; i<int(segs.size()); ++i )
		segs[i]->sampling = samp;
}

int        __cdecl QUB_IdlSegments::getClsCount()
{
	int i, maxcount = 0;

	for ( i=0; i<int(segs.size()); ++i )
		maxcount = max( maxcount, segs[i]->getClassCount() );
	return maxcount;
}
/*
	map<int, pair<double, double>> defaults;

	for ( i=0; i<segs.size(); ++i ) {
		int nclass = segs[i]->getClassCount();
		while ( nclass > maxcount ) {
			defaults[ maxcount ] = pair<double,double>( segs[i]->getAmp(maxcount), segs[i]->getStd(maxcount) );
			++maxcount;
		}
	}
	for ( i=0; i<segs.size(); ++i ) {
		int nclass = segs[i]->getClassCount();
		while ( nclass < maxcount )

}
*/


int        __cdecl QUB_IdlSegments::getSegCount()
{
	return segs.size();
}

QTR_Impl * __cdecl QUB_IdlSegments::_getSegmentation()
{
	QUB_Tree t = getSegmentation();
	QTR_INCREF( t.getImpl() );
	return t.getImpl();
}

QUB_Tree   __cdecl QUB_IdlSegments::getSegmentation()
{
	QUB_Tree segm = QUB_Tree::Create("");
	segm.setNumData( QTR_TYPE_INT, getSegCount(), 2, (void*) 0 );
	for ( int i = getSegCount() - 1; i >= 0; --i ) {
		segm.dataAs(i, 0, 0) = segs[i]->first;
		segm.dataAs(i, 1, 0) = segs[i]->last;
	}
	return segm;
}

QTR_Impl * __cdecl QUB_IdlSegments::_toTree()
{
	QUB_Tree t = toTree();
	QTR_INCREF( t.getImpl() );
	return t.getImpl();
}

QUB_Tree   __cdecl QUB_IdlSegments::toTree()
{
	if ( getSegCount() )
		return getDwells(segs[0]->first, segs[ segs.size()-1 ]->last, true, true);
	
	QUB_Tree ds = QUB_Tree::Create("DataSet");
	ds["sampling"].setData(QTR_TYPE_DOUBLE, sampling * 1.0e3);
	
	return ds;
}

inline int round( double x )
{
	return (int) floor( x + 0.5 );
}

int        __cdecl QUB_IdlSegments::readDWT(char *fileName)
{
	QUB_Tree dwt = QUB_Tree( QUB_Idl_ReadDWT(fileName) );
	if ( dwt.isNull() )
		return -1;
	QTR_DECREF( dwt.getImpl() );

	lock();

	double sampMS = dwt["sampling"].dataAsDouble( getSampling() * 1.0e3 );
	setSampling( sampMS * 1.0e-3 );

	int at = 0;
	for ( QUB_TreeMonoIter segi = dwt.find("Segment"); ! segi->isNull(); segi.nextSameName() ) {
		QUB_Tree seg = *segi;
		int ndwt = seg["DwellCount"].dataAsInt(0);
		float *tdwt = (float *) seg["Durations"].data();

		seg["Firsts"].setNumData(QTR_TYPE_INT, ndwt, 1, (void*)0);
		int *firsts = (int *) seg["Firsts"].data();
		seg["Lasts"].setNumData(QTR_TYPE_INT, ndwt, 1, (void*)0);
		int *lasts = (int *) seg["Lasts"].data();

		for ( int i=0; i<ndwt; ++i ) {
			firsts[i] = at;
			at += round( tdwt[i] / sampMS );
			lasts[i] = at - 1;
		}

		seg.setNumData(QTR_TYPE_INT, 2, 1, (void*) 0);
		seg.dataAs(0, 0, 0) = firsts[0];
		seg.dataAs(1, 0, 0) = lasts[ ndwt-1 ];

		addSeg( firsts[0], lasts[ ndwt-1 ] );
	}

	setDwells( dwt );

	unlock();

	return 0;
}


QTR_Impl * __cdecl QUB_IdlSegments::_getDwells(int f, int l, int fragments, int withDurs)
{
	QUB_Tree dwells = getDwells(f, l, fragments, withDurs);
	QTR_INCREF(dwells.getImpl());
	return dwells.getImpl();
}

QUB_Tree   __cdecl QUB_IdlSegments::getDwells(int f, int l, int fragments, int withDurs)
{
	QUB_Tree ds = QUB_Tree::Create("DataSet");
	ds["sampling"].setData(QTR_TYPE_DOUBLE, sampling * 1.0e3);

	getDwellsInto(f, l, fragments, withDurs, ds, 1);

	return ds;
}

void       __cdecl QUB_IdlSegments::_getDwellsInto(int f, int l, int fragments, int withDurs,
								  QTR_Impl *dataSet, int insertAsFirst )
{
	getDwellsInto(f, l, fragments, withDurs, QUB_Tree(dataSet), insertAsFirst);
}

void       __cdecl QUB_IdlSegments::getDwellsInto(int f, int l, int fragments, int withDurs,
								 QUB_Tree dataSet, int insertAsFirst )
{
	lock();

	int sf, sl;
	getSegRange(f, l, sf, sl);
	if ( sf == segs.size() )
		sf = sl = segs.size() - 1;
	if ( sf < 0 )
		return;

	QUB_Tree afterSeg;
	if ( ! insertAsFirst )
		afterSeg = * dataSet.rfind("Segment");

	for ( int i=sf; i<=sl; ++i ) {
		QUB_IdlSeg& seg = * segs[i];
		seg.getDwells( f, l, fragments, withDurs, dataSet, afterSeg );
	}

	unlock();
}

void       __cdecl QUB_IdlSegments::_getDwellsIntoSegs(QTR_Impl* dataSet, int withDurs)
{
	getDwellsIntoSegs( QUB_Tree(dataSet), withDurs );
}

void       __cdecl QUB_IdlSegments::getDwellsIntoSegs(QUB_Tree dataSet, int withDurs)
{
	int sf, sl;

	lock();

	for ( QUB_TreeMonoIter segi = dataSet.find("Segment"); ! segi->isNull(); segi.nextSameName() ) {
		getSegRange(segi->dataAs(0, 0, 0), segi->dataAs(1, 0, 0), sf, sl);
		if ( sf == segs.size() )
			sf = sl = segs.size() - 1;
		segs[sf]->getDwellsInto(*segi, withDurs);
	}

	unlock();
}

void       __cdecl QUB_IdlSegments::_getDwellsIntoWDrop(int f, int l, int withDurs,
								  int dropFirstLast, int dropIfCls, int dropCls,
								  QTR_Impl *dataSet, int insertAsFirst )
{
	getDwellsIntoWDrop( f, l, withDurs, dropFirstLast, dropIfCls, dropCls, QUB_Tree(dataSet), insertAsFirst );
}

void       __cdecl QUB_IdlSegments::getDwellsIntoWDrop(int f, int l, int withDurs,
								  int dropFirstLast, int dropIfCls, int dropCls,
								  QUB_Tree dataSet, int insertAsFirst )
{
	lock();

	int sf, sl;
	getSegRange(f, l, sf, sl);
	if ( sf == segs.size() )
		sf = sl = segs.size() - 1;

	QUB_Tree afterSeg;
	if ( ! insertAsFirst )
		afterSeg = * dataSet.rfind("Segment");

	for ( int i=sf; i<=sl; ++i ) {
		QUB_IdlSeg& seg = * segs[i];
		seg.getDwellsWDrop( f, l, withDurs, dropFirstLast, dropIfCls, dropCls, dataSet, afterSeg );
	}

	unlock();
}

int        __cdecl QUB_IdlSegments::_setDwells(QTR_Impl *dataSet)
{
	return setDwells( QUB_Tree(dataSet) );
}

int        __cdecl QUB_IdlSegments::setDwells(QUB_Tree dataSet)
{
	lock();
	QUB_TreeMonoIter segi = dataSet.find("Segment");
	int s = 0;

	while ( ! segi->isNull() ) {
		int f = segi->dataAsInt(0, 0);
		int l = segi->dataAsInt(-1, 1);

		while ( (s < int(segs.size())) && (segs[s]->last < f) )
			++s;
		while ( (s < int(segs.size())) && (segs[s]->first <= l) ) {
			if ( segs[s]->first >= f && segs[s]->last <= l )
				resetSeg(s);
			segs[s]->setDwells(*segi);
			++s;
		}

		--s; // for weird overlaps
		segi.nextSameName();
	}

	unlock();
	return 0;
}

int        __cdecl QUB_IdlSegments::_setDwellsSeg(QTR_Impl *seg)
{
	return setDwellsSeg( QUB_Tree(seg) );
}

int        __cdecl QUB_IdlSegments::setDwellsSeg(QUB_Tree seg)
{
	lock();

	int f = seg.dataAsInt(0, 0);
	int l = seg.dataAsInt(-1, 1);
	int s = 0;

	while ( (s < int(segs.size())) && (segs[s]->last < f) )
		++s;
	while ( (s < int(segs.size())) && (segs[s]->first <= l) ) {
		if ( segs[s]->first >= f && segs[s]->last <= l )
			resetSeg(s);
		segs[s]->setDwells(seg);
		++s;
	}

	unlock();
	return 0;
}

int        __cdecl QUB_IdlSegments::setDwells(int count, QUB_IdlDwell *buf)
{
	lock();
	int s = 0;

	int f = buf[0].first;
	int l = buf[count-1].last;

	while ( (s < int(segs.size())) && (segs[s]->last < f) )
		++s;
	while ( (s < int(segs.size())) && (segs[s]->first <= l) ) {
		if ( segs[s]->first >= f && segs[s]->last <= l )
			resetSeg(s);
		segs[s]->setDwells(count, buf);
		++s;
	}

	unlock();
	return 0;
}

int        __cdecl QUB_IdlSegments::setDwellsFLC(int count, int *ff, int *ll, int *cc)
{
	lock();
	int s = 0;

	int f = ff[0];
	int l = ll[count-1];

	while ( (s < int(segs.size())) && (segs[s]->last < f) )
		++s;
	while ( (s < int(segs.size())) && (segs[s]->first <= l) ) {
		if ( segs[s]->first >= f && segs[s]->last <= l )
			resetSeg(s);
		segs[s]->setDwells(count, ff, ll, cc);
		++s;
	}

	unlock();
	return 0;
}

int        __cdecl QUB_IdlSegments::clear()
{
	int rslt = 0;
	lock();
	if ( segs.size() )
		rslt = erase(segs[0]->first, segs[segs.size()-1]->last);
	for ( int i=0; i<int(segs.size()); ++i )
		segs[i]->setClassCount(0);
	unlock();
	return rslt;
}

int        __cdecl QUB_IdlSegments::addSeg(int f, int l)
{
	lock();
	int sf, sl, s;
	getSegRange(f, l, sf, sl);

	if ( sf >= int(segs.size()) ) {
		segs.push_back( new QUB_IdlSeg(freezer, chunkStorage, f, l, sampling) );
		unlock();
		return 0;
	}

	if ( segs[sf]->first < f ) {
		breakSeg(f);
		++sf;
	}
	
	vector<QUB_IdlSeg*>::iterator segi = segs.begin();
	for ( s=0; s<sf; ++s )
		++segi;
	segs.insert( segi, new QUB_IdlSeg(freezer, chunkStorage, f, l, sampling) );

	for ( s=sf+1; s<int(segs.size()); ++s ) {
		segs[s]->offset(l-f+1);
	}

	unlock();
	return 0;
}

int        __cdecl QUB_IdlSegments::removePoints(int f, int l)
{
	lock();
	int sf, sl, s;
	getSegRange(f, l, sf, sl);
	int count = l - f + 1;

	vector<QUB_IdlSeg*>::iterator segi = segs.begin();
	for ( s=0; s<sf; ++s )
		++ segi;

	for ( s=sf; s<=sl; ++s, ++segi ) {
		QUB_IdlSeg& seg = * segs[s];
		bool hasBefore = seg.first < f;
		bool hasAfter = l < seg.last;

		if ( hasBefore && hasAfter ) {
			breakSeg(f);
			segs[s+1]->setFirst(l+1);
		}
		else if ( hasBefore ) {
			seg.setLast(f-1);
		}
		else if ( hasAfter ) {
			seg.setFirst(l+1);
			--sl;
		}
		else {
			if ( segs.size() > 1 ) {
				delete segs[s];
				segs.erase(segi);
				if ( s )
					--segi;
				--s;
				--sl;
			}
			else {
				segs[s]->setLast( segs[s]->first - 1 );
			}
		}
	}

	for ( s=sl+1; s<int(segs.size()); ++s ) {
		segs[s]->offset( - count );
	}

	unlock();
	return 0;
}

int        __cdecl QUB_IdlSegments::insertPoints(int f, int l)
{
	lock();
	int sf, s;
	getSegRange(f, f, sf, sf);

	int nseg = segs.size();
	if ( (sf == nseg) && (segs[nseg-1]->last == (f-1)) )
		sf = nseg-1;

	int rslt = 0;
	if ( sf == int(segs.size()) )
		rslt = addSeg(f, l);
	else {
		segs[sf]->insert(f, l);
		for ( s=sf+1; s<int(segs.size()); ++s )
			segs[s]->offset(l - f + 1);
	}
	unlock();
	return rslt;
}

int        __cdecl QUB_IdlSegments::breakSeg(int at)
{
	lock();

	int sf, s;
	getSegRange(at, at, sf, sf);
	int rslt = 0;
	
	if ( sf == int(segs.size()) )
		rslt = -1;
	else if ( segs[sf]->first == at )
		rslt = 1;
	else {
		vector<QUB_IdlSeg*>::iterator segi = segs.begin();
		for ( s=0; s<sf; ++s )
			++segi;
		
		int segfirst = segs[s]->first;
		segs.insert(segi, new QUB_IdlSeg(freezer, chunkStorage, segfirst, at-1, sampling));
		vector<QUB_IdlDwell> dwells;
		if ( true ) {
			QUB_IdlMaps::iterator di = segs[sf+1]->begin();
			while ( di && (di->first + segfirst) <= at ) {
				dwells.push_back( *di );
				++di;
			}
		}
		segs[sf+1]->setFirst( at );
		segs[sf]->maps.set(dwells.size(), &(dwells[0]), 0);
		segs[sf]->ampStd = segs[sf+1]->ampStd;
	}
	unlock();
	return rslt;
}

int        __cdecl QUB_IdlSegments::joinSegs(int at)
{
	lock();

	int s, sf, sl;
	getSegRange(at-1, at, sf, sl);
	int rslt = 0;

	if ( sf == int(segs.size()) )
		rslt = -1;
	if ( sf == sl )
		rslt = -2;
	else {
		vector<QUB_IdlSeg*>::iterator segi = segs.begin();
		for ( s=0; s<sl; ++s )
			++segi;

		QUB_IdlSeg& segA = * segs[sf];
		int oldApoints = segA.last - segA.first + 1;
		vector<QUB_IdlDwell> dwells;
		if ( true ) {
			QUB_IdlMaps::iterator di = segs[sl]->begin();
			while ( di ) {
				dwells.push_back( *di );
				++di;
			}
		}

		segA.setLast( segs[sl]->last );
		segA.maps.set(dwells.size(), &(dwells[0]), - oldApoints);
		if ( segA.getClassCount() < segs[sl]->getClassCount() )
			segA.ampStd = segs[sl]->ampStd;

		delete segs[sl];
		segs.erase(segi);
	}
	unlock();
	return rslt;
}

int        __cdecl QUB_IdlSegments::erase(int f, int l)
{
	if ( f > l ) return 0;
	lock();
	int s, sf, sl, ff, ll;
	getSegRange(f, l, sf, sl);
	for ( s=sf; s<=sl; ++s ) {
		QUB_IdlSeg& seg = * segs[s];
		ff = max(f, seg.first);
		ll = min(l, seg.last);
		if ( seg.first == ff && seg.last == ll )
			resetSeg(s);
		else if ( ff <= ll )
			seg.set(ff, ll, -1);
	}
	unlock();
	return 0;
}

int        __cdecl QUB_IdlSegments::join(int f, int l)
{
	lock();
	int s, sf, sl, ff, ll;
	getSegRange(f, l, sf, sl);
	for ( s=sf; s<=sl; ++s ) {
		QUB_IdlSeg& seg = * segs[s];
		ff = max(f, seg.first);
		ll = min(l, seg.last);
		if ( ff <= ll )
			seg.join(ff, ll);
	}
	unlock();
	return 0;
}

QUB_IdlSegments::iterator QUB_IdlSegments::begin()
{
	if ( segs.size() )
		return iterator(*this, 0, segs[0]->begin());
	return iterator(*this, -1, QUB_IdlMaps::iterator());
}

QUB_IdlSegments::iterator QUB_IdlSegments::end()
{
	if ( segs.size() )
		return iterator(*this, segs.size()-1, segs[segs.size()-1]->end());
	return iterator(*this, -1, QUB_IdlMaps::iterator());
}

QUB_IdlSegments::iterator QUB_IdlSegments::find(int point)
{
	for ( int i=0; i<int(segs.size()); ++i )
		if ( segs[i]->first <= point && point <= segs[i]->last )
			return iterator(*this, i, segs[i]->find(point));
	return iterator(*this, -1, QUB_IdlMaps::iterator());
}


QUB_IdlSegmentsIter_iface* __cdecl QUB_IdlSegments::_begin()
{
	return begin().copy();
}

QUB_IdlSegmentsIter_iface* __cdecl QUB_IdlSegments::_end()
{
	return end().copy();
}

QUB_IdlSegmentsIter_iface* __cdecl QUB_IdlSegments::_find(int point)
{
	return find(point).copy();
}

int        __cdecl QUB_IdlSegments::useTempFile(int doUse)
{
	if ( doUse ) {
		chunkStorage.getRoot().saveAsTemp();
		return chunkStorage.getRoot().inFile() ? 0 : -1;
	}
	else {
		chunkStorage.getRoot().close();
		return 0;
	}
}

int        __cdecl QUB_IdlSegments::saveASCII(char *fileName)
{
	ofstream out;
	out.open(fileName);
	if ( ! out )
		return -1;

	int f, l, c;
	double a;
	iterator di = begin();
	while ( di ) {
		di.read(&f, &l, &c, &a);
		double dur = sampling * 1e3 * (l - f + 1);
		for ( int i=(l - f + 1); i>0; --i )
			out << a << "\t" << dur << "\r\n";
		++di;
	}
	
	return 0;
}


extern "C" MIL_API QUB_IdlSegments_iface *    __stdcall QUB_IdlSegments_Create( double sampling )
{
	return new QUB_IdlSegments(sampling);
}


//-----
extern "C" MIL_API QTR_Impl * __stdcall QUB_Idl_ReadDWT(char *filename){
	QUB_Tree dwt = QUB_Tree::Create("DataSet");

	dwt["sampling"].setNumData(QTR_TYPE_DOUBLE, 1, 1, (void*)0);
	double& samp = * (double *) dwt["sampling"].data();
	samp = 1;
	bool noSamplingGiven = true;

	QUB_Tree seg;

	ifstream in;
	in.open(filename);
	if ( ! in )
		return NULL;

	int len;
	char * line;
	char * pchBuf, * pTok;

	while ( ! in.eof() ) {
		int i;
		int ndwell = 0;
		int nclass = 0;
		vector<double> amp, sd;
		double start, at = 0.0;
		istream_linereader rdr;

		//----------------------------- Following section rewritten 8-2004 JTB
		//----- Find a 'Segment:'
		do { 
			line = rdr.readLine(in, len);

			if( memcmp("Sampling duration (ms):",line,23)== 0 	// ICE : sampling.
					&& atof(line+24) > 0.0 ) {
				samp=atof(line+24);
				noSamplingGiven = false;
				}

			} while( ! in.eof() && memcmp(line,"Segment:", 8)!=0  );

		//----- copy read line to buffer
		pchBuf = new char[strlen(line)+1];
		strcpy(pchBuf,line);

		//----- Parse 'tokens' from input line
		pTok=strtok(pchBuf," \t");
		nclass = 0;
		start = at;

		while( pTok=strtok(NULL," \t")) { // Ok to lose 1st token - it is always Segment:
			if( strcmp(pTok, "Dwells:") == 0 )			// QUB & ICE : Dwells:
				if( pTok=strtok(NULL," ") )
					ndwell=atoi(pTok);

			if ( strcmp(pTok, "Sampling(ms):") == 0 )	// QUB : Sampling(ms):
				if( pTok=strtok(NULL," ") ) {
					samp=atof(pTok);
					noSamplingGiven = false;
					}

			if ( strcmp(pTok, "Start(ms):") == 0 )		// QUB : Start(ms):
				if( pTok=strtok(NULL," ") )
					start=max(at,atof(pTok));

			if ( strcmp(pTok, "ClassCount:") == 0 )		// QUB : ClassCount: (list of pairs of amp,sd)
				if( pTok=strtok(NULL," ") ) { 
					nclass=atoi(pTok);
					amp.resize( nclass );
					sd.resize( nclass );
					for ( i=0; i<nclass && pTok!=NULL ; ++i ) { 
						if( pTok=strtok(NULL," ") )
							amp[i]=atof(pTok);
						if( pTok=strtok(NULL," ") )
							sd[i]=atof(pTok);
						}
					}

			if( strcmp(pTok, "Starting") == 0			// ICE : Starting time (ms):
					&& (pTok=strtok(NULL," "))
					&& strcmp(pTok, "time") == 0
					&& (pTok=strtok(NULL," "))
					&& strcmp(pTok, "(ms):") == 0
					&& (pTok=strtok(NULL," ")) )
				start=max(at,atof(pTok));

			if( strcmp(pTok, "Amp:") == 0 )				// ICE : Amp: (list of amps)
				while ( (pTok=strtok(NULL," "))!=NULL && *pTok < 'A' ) { 
					++nclass;
					amp.resize( nclass );
					sd.resize( nclass );
					amp[nclass-1]=atof(pTok);
					sd[nclass-1]=0.0;
					}
			}
		delete[] pchBuf;

		//=========================================================================
		/*  Changed to code above 8-2004 JTB  - for ICE compatibility.
		int len;
		char *line = rdr.readLine(in, len);
		while ( (ndwell == 0) && ! in.eof() ) {
			istrstream lineIn(line);
			lineIn.get(s, 16, ' ') >> i >> ws;
			lineIn.get(s, 16, ' ');
			if ( (strcmp(s, "Dwells:") == 0) && ! lineIn.eof() )
				lineIn >> ndwell >> ws;

			nclass = 0;
			start = at;
			while ( (nclass == 0) && ! lineIn.eof() ) {
				lineIn.get(s, 16, ' ');
				if ( strcmp(s, "Sampling(ms):") == 0 ) {
					lineIn >> samp >> ws;
					noSamplingGiven = false;
				}
				else if ( strcmp(s, "Start(ms):") == 0 )
					lineIn >> start >> ws;
				else if ( strcmp(s, "ClassCount:") == 0 ) {
					lineIn >> nclass >> ws;

					amp.resize( nclass );
					sd.resize( nclass );
					for ( i=0; lineIn && i<nclass; ++i )
						lineIn >> amp[i] >> sd[i];
				}
				else
					lineIn >> d >> ws;
			}
			if ( start < at )
				start = at;

			if ( ndwell == 0 )
				line = rdr.readLine(in, len);
		}
		*/

		//-------------------------------------------------------------------
		if ( ndwell ) {
			seg = dwt.insertChild( seg, "Segment" );
			seg["start"].setData(QTR_TYPE_DOUBLE, start);
			seg["Classes"].setNumData(QTR_TYPE_INT, ndwell, 1, (void*)0);
			seg["Durations"].setNumData(QTR_TYPE_FLOAT, ndwell, 1, (void*)0);
			int *idwt = (int *) seg["Classes"].data();
			float *tdwt = (float *) seg["Durations"].data();

			set<int> classesPresent;

			for ( i=0; in && i<ndwell; ++i ) {
				in >> idwt[i] >> tdwt[i];
				classesPresent.insert(idwt[i]);
				at += tdwt[i];
				if ( noSamplingGiven && (tdwt[i] < samp) )
					samp = tdwt[i];
			}
			ndwell = i; // drop events that couldn't be read

			// tune up classes: add amp, sd for missing high-numbered; renumber negative as highest number, neg amp
			int minCls = 0, maxCls = nclass - 1;
			for ( set<int>::iterator ci = classesPresent.begin();
				  ci != classesPresent.end();
				  ++ci ) {
				if ( *ci < minCls )
					minCls = *ci;
				if ( *ci > maxCls )
					maxCls = *ci;
			}
			while ( nclass <= maxCls ) {
				amp.push_back( (double) nclass );
				sd.push_back( (double) 0.1 );
				++nclass;
			}
			if ( minCls < 0 ) {
				int offset = maxCls - minCls + 1;
				for ( int i=0; i<ndwell; ++i ) {
					if ( idwt[i] < 0 )
						idwt[i] += offset;
				}
				while ( nclass < offset ) {
					amp.push_back( (double) (nclass - offset) );
					sd.push_back( (double) 0.1 );
					++nclass;
				}
			}

			seg["DwellCount"].setData(QTR_TYPE_INT, ndwell);
			seg["amp"].setNumData(QTR_TYPE_DOUBLE, nclass, 1, &(amp[0]));
			seg["sd"].setNumData(QTR_TYPE_DOUBLE, nclass, 1, &(sd[0]));
		}
	}
	
	// HEKA (and others?) give event durations in seconds not milliseconds, and
	// QUB doesn't deal well with gigahertz sampling
	if ( noSamplingGiven && (samp < 1e-3/*megahertz*/) ) {
		samp *= 1.0e3;

		for ( QUB_TreeMonoIter segi = dwt.find("Segment"); ! segi->isNull(); segi.nextSameName() ) {
			int ndwt = (*segi)["DwellCount"].dataAsInt();
			float *tdwt = (float *) (*segi)["Durations"].data();
			for ( int i=0; i<ndwt; ++i )
				tdwt[i] *= 1.0e3;
		}
	}
	
	QTR_INCREF( dwt.getImpl() );
	return dwt.getImpl();
}


typedef struct
{
  // short epi; // clampex files only
  int start; // samples
  float amp; 
  int dur; // samples
  float std;
  short cls;
  short notes; // bitmap: {0: too short; 1: amp manually adjusted; 2: rejected}
  float baseline; // cls 0 only
} evl_event;

#define EVL_CLOSED_BYTES 24
#define EVL_OPEN_BYTES   20


extern "C" MIL_API QTR_Impl * __stdcall QUB_Idl_ReadEVL(char *filename)
{
	FILE *f = fopen(filename, "rb");
	if ( ! f )
		return NULL;
  
	fseek(f, 0, SEEK_END);
	int evlen = ftell(f);
  
	fseek(f, 158, SEEK_SET);
	float sampling;
	fread(&sampling, sizeof(float), 1, f);
	sampling /= 1000.0;
  
	fseek(f, 186, SEEK_SET);
	short nclass;
	fread(&nclass, sizeof(short), 1, f);
	vector<double> amp(nclass, 0.0), std(nclass, 0.0);

	evl_event evt;
	int ndwt = 0;
	vector<int> idwt;
	vector<float> tdwt;

	int at = 256;
	int cur_sample = 0;
	float start;

	int err;
  
	while ( at < evlen ) {
		err = fseek(f, at, SEEK_SET);
		// cerr << "seeking " << at << " (" << err << ")" << "\r\n";
    
		if ( EVL_OPEN_BYTES == (err = fread(&evt, 1, EVL_OPEN_BYTES, f)) ) {
			// cerr << "read " << err << " bytes" << "\r\n";
			// start amp dur std cls
			if ( ndwt == 0 )
				start = evt.start * sampling;
			  /*
			  cerr << "\tstart:\t" << evt.start << "\r\n";
			  cerr << "\tamp:\t" << evt.amp << "\r\n";
			  cerr << "\tdur:\t" << evt.dur << "\r\n";
			  cerr << "\tstd:\t" << evt.std << "\r\n";
			  cerr << "\tcls:\t" << evt.cls << "\r\n";
			  cerr << "\tnotes:\t" << evt.notes << "\r\n";
			  cerr << "\r\n";
			  */
			// QUB-style missed events: append "missing" time to prev.event:
			if ( ndwt && (evt.start > cur_sample) ) {
				tdwt[ndwt-1] += sampling * (evt.start - cur_sample);
				cur_sample = evt.start;
			}
			/*
			if ( ndwt && (evt.start > cur_sample) ) {
				cerr << "breaking segment (" << evt.start << " > " << cur_sample << ")" << "\r\n";
				printSeg(nclass, amp, std, ndwt, idwt, tdwt, sampling, start);
			}
			*/
			++ndwt;
			idwt.push_back( evt.cls );
			tdwt.push_back( sampling * evt.dur );
			amp[ evt.cls ] += evt.amp;
			std[ evt.cls ] += evt.std;
			cur_sample = evt.start + evt.dur;

			at += evt.cls ? EVL_OPEN_BYTES : EVL_CLOSED_BYTES;
		}
		else {
			break;
		}
	}
  
	fclose(f);

	
	QUB_Tree dwt = QUB_Tree::Create("DataSet");

	if ( ndwt ) {
		for ( int i=0; i<nclass; ++i ) {
			amp[i] /= ndwt;
			std[i] /= ndwt;
		}
			
		dwt["sampling"].setData( QTR_TYPE_DOUBLE, (double) sampling );

		QUB_Tree seg = dwt["Segment"];
		seg["start"].setData(QTR_TYPE_DOUBLE, start);
		seg["amp"].setNumData(QTR_TYPE_DOUBLE, nclass, 1, &(amp[0]));
		seg["sd"].setNumData(QTR_TYPE_DOUBLE, nclass, 1, &(std[0]));
		seg["DwellCount"].setData(QTR_TYPE_INT, ndwt);
		seg["Classes"].setNumData(QTR_TYPE_INT, ndwt, 1, &(idwt[0]));
		seg["Durations"].setNumData(QTR_TYPE_FLOAT, ndwt, 1, &(tdwt[0]));
	}
  
	QTR_INCREF( dwt.getImpl() );
	return dwt.getImpl();
}


void ReadTAC_InsertSeg(QUB_Tree dwt, QUB_Tree& seg, double start_ms, vector<double>& amps, vector<int>& idwt, vector<float>& tdwt)
{
	seg = dwt.insertChild(seg, "Segment");
	seg["start"].setData(QTR_TYPE_DOUBLE, start_ms);
	seg["amp"].setNumData(QTR_TYPE_DOUBLE, amps.size(), 1, &(amps[0]));
	seg["DwellCount"].setData(QTR_TYPE_INT, idwt.size());
	seg["Classes"].setNumData(QTR_TYPE_INT, idwt.size(), 1, &(idwt[0]));
	seg["Durations"].setNumData(QTR_TYPE_FLOAT, tdwt.size(), 1, &(tdwt[0]));
}

extern "C" MIL_API QTR_Impl * __stdcall QUB_Idl_ReadTAC(char *filename)
{
	ifstream in;
	in.open(filename);
	if ( ! in )
		return NULL;

	istream_linereader rdr;
	int len;
	char * line;

	vector<int> idwt;
	vector<float> tdwt;
	vector<double> amps;
	int seg_ix = -1;
	double last_secs = 0.0;
	int last_level = 0;
	double smallest_dt = 1.0e5;
	double start_ms = 0.0;
	double seg_ms = 0.0;

	QUB_Tree dwt = QUB_Tree::Create("DataSet");
	QUB_Tree seg;

	while ( ! in.eof() ) {
		line = rdr.readLine(in, len);
		if ( (! len) || (line[0] < '0') || (line[0] > '9') )
			continue;

		istringstream fromline(line);
		int sweep;
		int level = -1; // to ensure it's a full line, make sure this is overwritten with non-neg
		double secs, preamp, amp, tag;
		fromline >> sweep >> secs >> preamp >> amp >> level >> tag;
		if ( level < 0 )
			continue;

		if ( seg_ix != sweep ) {
			if ( idwt.size() ) {
				ReadTAC_InsertSeg(dwt, seg, start_ms, amps, idwt, tdwt);
				idwt.clear();
				tdwt.clear();
				start_ms += seg_ms;
				last_level = 0;
			}
			seg_ix = sweep;
		}

		if ( secs < last_secs )
			last_secs = 0.0;
		double dt = 1e3 * (secs - last_secs);
		last_secs = secs;
		seg_ms += dt;
		if ( dt < smallest_dt )
			smallest_dt = dt;

		idwt.push_back(last_level);
		tdwt.push_back((float) dt);
		if ( amps.size() <= level )
			amps.resize(level+1);
		amps[level] = amp * 1e12;
		last_level = level;
	}
	if ( idwt.size() ) {
		idwt.push_back(last_level);
		tdwt.push_back((float) smallest_dt);
		ReadTAC_InsertSeg(dwt, seg, start_ms, amps, idwt, tdwt);
	}
	// first event has unknown duration -- set all segs' first duration to smallest encountered:
	for (QUB_TreeMonoIter segi = dwt.find("Segment"); ! segi->isNull(); segi.nextSameName())
		((float*)(*segi)["Durations"].data())[0] = (float) smallest_dt;
	smallest_dt /= 12;
	dwt["sampling"].setData(QTR_TYPE_DOUBLE, smallest_dt);

	QTR_INCREF( dwt.getImpl() );
	return dwt.getImpl();
}


